#include <string.h>
#include "owbasic.h"
#include "comp.h"

/*
   Compiler list management

   The list of symbols has the following structure:


   Each entry contains
   - xdata[n]   type
   - xdata[n+1] extended type (DTYPEDEF/DPROCFUNC: return value type)
   - xdata[n+2] entry length


   The usage of the following entries differs:
     Variables:
     - xdata[n+3] runtime address (DREFERENCE: DLABEL entry)
     - xdata[n+4] element number total (DREFERENCE: pointer to reference position)
     - xdata[n+5] element number per line
     - xdata[n+6] element size
     - xdata[n+7] ...
         identifier (character string, finished with '\0')
     - xdata[n+len-1] n -> pointer to start of entry

     DLABEL:
     - xdata[n+3] code address (0 - not defined yet but referenced)
     - xdata[n+4] -
     - xdata[n+5] -
     - xdata[n+6] -
     - xdata[n+7] ...
         identifier (character string, finished with '\0')
     - xdata[n+len-1] n -> pointer to start of entry

     DREFERENCE:
     - xdata[n+3] DLABEL entry address
     - xdata[n+4] pointer to reference position
     - xdata[n+5] -
     - xdata[n+6] -
     - xdata[n+7] 0 -> no name, empty strong
     - xdata[n+8] n -> pointer to start of entry

     DPROCFUNC:
     - xdata[n+3] procedure address
     - xdata[n+4] length of code for overlay procedures, otherwise 0
     - xdata[n+5] default parameter values address, 0 for no default parameters
     - xdata[n+6] extended datatypes address, 0 for no extended datatypes
     - xdata[n+7] ...
         identifier (character string, finished with '\0')
     - xdata[n+len-1] n -> pointer to start of entry

     DCONSTANT:
     - xdata[n+3] constant value
     - xdata[n+4] constant datatype
     - xdata[n+5] extended datatype
     - xdata[n+6] -
     - xdata[n+7] ...
         identifier (character string, finished with '\0')
     - xdata[n+len-1] n -> pointer to start of entry

     DTYPEDEF:
     - xdata[n+3] DFPTR: DPROCFUNC definition address / -
     - xdata[n+4] -
     - xdata[n+5] -
     - xdata[n+6] -
     - xdata[n+7] ...
         identifier (character string, finished with '\0')
     - xdata[n+len-1] n -> pointer to start of entry



   Constants management:
     constant values for integers and floats
*/

#define LIST_ADDROFFSET

#define LIST_TYPE(a)      xdata[(a) - 3]
#define LIST_EXTTYPE(a)   xdata[(a) - 2]
#define LIST_ENTRYLEN(a)  xdata[(a) - 1]
#define LIST_ADDRESS(a)   xdata[a]
#define LIST_ARRSIZE(a)   xdata[(a) + 1]
#define LIST_ARR2SIZE(a)  xdata[(a) + 2]
#define LIST_ELEMSIZE(a)  xdata[(a) + 3]
#define LIST_NAME(a)      ((uchar FAR*) (&xdata[(a) + 4]))

int listnext; /* next entry (compile time) */

/* int datanext; next useable address (for runtime), now defined in bin.c */
/* int firstconst; first address used for constants, now defined in bin.c */

int lastaddr; /* address of the last created entry */


int search_procfunc (uchar* name, type_t rtype)
{
  datatype_t type;

  type.type = DPROCFUNC;
  type.ext = rtype;
  return (searchentry_name (&type, name));
}

int new_procfunc (uchar* name, PC pa, type_t rtype)
{
  register int addr;
  datatype_t type;

  type.type = DPROCFUNC;
  type.ext = rtype;

  addr = newentry_custom (type, name, (MAXPARAM + 1) * sizeof (type_t) / sizeof (int));
  xdata[addr] = (int) (pa.c);
  xdata[addr + 1] = xdata[addr + 2] = xdata[addr + 3] = 0;
  return (addr);
}

int FAR* custom_data (int addr)
{
  return (&xdata[xdata[addr - 3 + xdata[addr - 1] - 2]]);
}

int newentry_custom (datatype_t dtype, const uchar* name, int adsize)
{
  int beg, addr, next;

  addr = newentry (dtype, name);
  beg = addr - 3 + xdata[addr - 1] - 1;

  ++adsize; /* add space for pointer to begin of additional data */
  next = addr - 3 + xdata[addr - 1] + adsize;
  xdata[addr - 1] += adsize;
  xdata[next - 1] = addr - 3;
  xdata[next - 2] = beg;
  listnext += adsize;
  xdata[listnext] = 0;
  return (addr);
}

int newentry (datatype_t dtype, const uchar* name)
{
  /* warning: length of name is max(IDENTLEN,NAMELEN+1)*/
  int adr, namelength;
  int datasize;
  int headersize;
  uchar FAR* dptr;
  const uchar* sptr;

  if (listnext > firstconst - (NAMELEN + 1 + 1) / sizeof (int) - 11) /* check free symbol space */
    serror (5);
  adr = listnext;                         /* position of this entry */
  sptr = name;                            /* address of identifier */
  dptr = (uchar FAR *) (&xdata[adr + 7]); /* destination address */
  while (*sptr)
    *(dptr++) = *(sptr++); /* copy identifier */
  *(dptr++) = 0;                          /* null as delimiter */
  namelength = strlen ((char*) name) + 1; /* number of bytes for name */
  namelength = (namelength + sizeof(int) - 1) / sizeof (int); /* string length in int */

  /*LibPutMsgDlg ("%s:%d", name, dtype.type);*/
  /* determine datasize (in sizeof (int))*/
  datasize = headersize = 1;
  if ((dtype.type & TYPEMASK) > LASTVARTYPE)
    datasize = headersize = 0;
  else if ((dtype.type & TYPEMASK) == DFLOAT)
    datasize = FLOATSIZE;

#if FALSE
  if (dtype & KREFERENCE)
  {
    datasize = 1; /*headersize = 2;*/ /* pointer and element count */
  }
  else
    switch (dtype & TYPEMASK)
    {
    case DENUM:    /* the value itself + len values */
    case DFPTR:    /* the value itself + len values */
    case DINTEGER: /* the value itself + len values */
    case DBOOLEAN: /* the value itself + len values */
    case DCHAR:    /* the value itself + len values */
    case DSTRING:  /* a pointer to the string + len values */
      datasize = 1; /*headersize = 2;*/
      break;
    case DFLOAT:  /* FLOATT size + len values */
      datasize = FLOATSIZE; /*headersize = 2;*/
      break;
    case DCONSTANT:  /* no runtime length, only at compiletime */
    case DLABEL:
    case DREFERENCE:
    default:
      datasize = 0; headersize = 0;
      break;
    }
#endif

  lastaddr = adr + 3;                /* address of last entry */

  /* allocate memory */
  if (dtype.type & KCONST)
    xdata[lastaddr] = firstconst;
  else
  {
    if ((dtype.type & TYPEMASK) == DLABEL)
      xdata[lastaddr] = 0; /* new label */
    else
    /* address of data at runtime
     * pointer points direct to data after header
     */
    {
      datanext += headersize; /* reserve header */
      xdata[lastaddr] = datanext; /* store runtime address of data */
      datanext += datasize;  /* reserve data space */
    }
  }

  if (datanext >= firstconst)
    serror (33); /* not enough space */

  xdata[lastaddr + 1] = 1;        /* number of elements (total) */
  xdata[lastaddr + 2] = 0;        /* number of elements per line */
  /* 0 as alement count causes less header size */
  xdata[lastaddr + 3] = datasize; /* size of *one* element */
  xdata[adr] = dtype.type;        /* store type of entry */
  xdata[adr + 1] = dtype.ext;        /* extended type */
  xdata[adr + 2] = namelength + 8;   /* store length of entry */
  xdata[adr + 7 + namelength] = adr; /* last value points to start of entry */
  listnext += namelength + 8;        /* place of next entry */
  xdata[listnext] = 0;               /* end of list */
  return (lastaddr);
}

int expandentry (int addr, int el, int el_per_line)
   /* expand size of runtime memory (arrays) */
{

  if (addr != lastaddr)
  {
    /*LibPutMsgDlg ("addr:%d\rlastaddr:%d", addr, lastaddr);
    LibPutMsgDlg ("name of addr:\r%s", Near ((char FAR*) (&xdata[addr + 4])));
    LibPutMsgDlg ("name of lastaddr:\r%s", Near ((char FAR*) (&xdata[lastaddr + 4])));*/
    serror (31);  /* only possible for last entry */
  }
  xdata[addr + 1] = el;                /* number of elements of array */
  xdata[addr + 2] = el_per_line;       /* number of elements per row */
  if (el_per_line != 0)
  {
    ++xdata[addr]; /* increase runtime address because of header */
    ++datanext;
  }
  datanext += (el - 1) * xdata[addr + 3];  /* allocate memory */
  if (datanext >= firstconst)
    serror (33); /* not enough space */
  return (0);
}

int searchentry_token (datatype_t* dtype)
{
  uchar n[IDENTLEN + 1];
  tokncpy (n, IDENTLEN + 1);
  searchentry_name (dtype, n);
}

int searchentry_name (datatype_t* dtype, const uchar* name)
{
  uchar n[IDENTLEN + 1];
  int ptr;
  int varlen;
  datatype_t vartype;
  uchar *p1;
  uchar FAR *pf;
  datatype_t rtype = *dtype;

  rtype.type &= TYPEMASK;
  ptr = 0; /* start search at beginning */
  /* copy and normalize name */
  strcpy_name (n, name);
  strlower (n); /* normalize name (lower case) */

  while (xdata[ptr] != 0) /* search entry */
  {
    vartype.type = xdata[ptr] & TYPEMASK;
    vartype.ext = xdata[ptr + 1];
    varlen = xdata[ptr + 2];

    if ((((rtype.type == DUNKNOWN) && (vartype.type <= LASTCVARTYPE)) /* search type unknown -> OK */
     || (vartype.type == rtype.type)) /* check given type */
     && ((rtype.ext == 0)         /* extended type unknown -> OK */
     || (rtype.ext == vartype.ext))) /* check extended type */
    {
      p1 = n;
      pf = (uchar FAR*) &xdata[ptr + 7];
      while ((*p1) && (*p1 == *pf))
      { /* compare names */
        ++pf;
        ++p1;
      }
      if (*p1 == *pf) /* name found */
      {
        *dtype = vartype;
        return (ptr + 3);
      }
    }

    ptr += varlen; /* next entry */
  }
  return (-1);
}

int findentry_token (datatype_t* dtype)
{
  uchar n[IDENTLEN + 1];
  tokncpy (n, IDENTLEN + 1);
  findentry_name (dtype, n);
}

int findentry_name (datatype_t* dtype, const uchar* name)
{
  uchar n[IDENTLEN + 1];
  int var;

  var = searchentry_name (dtype, name);
  if (var < 0) /* if not found */
  {
    if (declaration_required && ((dtype->type & TYPEMASK) <= LASTVARTYPE))
      serror (24); /* undefined variable */
    /* copy and normalize name */
    strcpy_name (n, name);
    strlower (n); /* normalize name (lower case) */

    dtype->type = DT (dtype->type);
    var = newentry (*dtype, n);
  }
  return (var);
}

#ifdef DEBUGLIST
void listshow (void)
{
  int ptr;
  uchar FAR *hptr;
  uchar t, elength, l;
  for (l = 0; l < 15; l++)
    OutTrm ('\r');
  ptr = 0;
  while (xdata[ptr] != 0)
  {
    t = xdata[ptr];
    elength = xdata[ptr + 2];
    hptr = (uchar FAR*) &(xdata[ptr + 7]);
    while (*hptr != 0)
      OutTrm (*(hptr++));
    OutTrm (' ');
    OutputITrm (t);
    OutputITrm (xdata[ptr + 1]);
    OutputITrm (xdata[ptr + 3]);
    OutputITrm (xdata[ptr + 4]);
    OutputITrm (xdata[ptr + 5]);
    OutTrm ('\r');
    ptr += elength;
  }
}
#endif

void listreset (void)
{
  listnext = 0;
  datanext = 0;
  xdata[0] = 0;
  firstconst = DATASIZE;
}

void listhide (int pos)
{
  /* removes local entries in symboltable logically
     labels and references are removed
     the name of variables is reduced to an emptystring and the
     entry is compacted */
  int sptr = pos; /* from .. */
  int dptr = pos; /* to .. */
  type_t typ;
  int len;

  check_labels (); /* link labels */

  while ((typ = xdata[sptr]) != 0)
  {
    /*typ = xdata[sptr];*/
    len = xdata[sptr + 2];
    if ((typ & TYPEMASK) > LASTVARTYPE)
      sptr += len; /* ignore (delete) entry*/
    else
    {
      xdata[dptr]     = typ; /* copy entry to new position */
      xdata[dptr + 1] = xdata[sptr + 1];
      xdata[dptr + 2] = 9; /* size */
      xdata[dptr + 3] = xdata[sptr + 3];
      xdata[dptr + 4] = xdata[sptr + 4];
      xdata[dptr + 5] = xdata[sptr + 5];
      xdata[dptr + 6] = xdata[sptr + 6];
      xdata[dptr + 7] = 0; /* no name */
      xdata[dptr + 8] = dptr;

      dptr += 9;
      sptr += len;
    }
    listnext = dptr; /* set pointer for next free place */
  }
  xdata[dptr] = 0; /* mark end */
}

void check_labels (void)
{
  int ptr;
  type_t t;
  int l;
  int maddr;
  uchar label[80];
  PC baddr, raddr;

#ifdef DEBUGLIST
  listshow ();
#endif

  ptr = 0;
  while (xdata[ptr] != 0)
  {
    t = xdata[ptr];
    l = xdata[ptr + 2];
    if ((t & TYPEMASK) == DREFERENCE)
    {
      maddr = xdata[ptr + 3];/* pointer to label data */
      baddr.p = (codet**) xdata[ptr + 4]; /* where to put address */
      raddr.c = (codet*)  xdata[maddr]; /* label address */
      if (raddr.c == NULL) /* label undefined */
      {
        /* label name at xdata[ptr + 4] */
        strcpy_nf (label,(uchar FAR*) &(xdata[maddr + 4]));
        serror1 (6054, label);
      }
      if (*baddr.p)
      {
        LINKJMP (baddr.i, raddr.c);
      }
      else
        *baddr.p = raddr.c;
    }
    ptr += l;
  }
}

void create_variables (void)
{
  int aentry;       /* pointer for entries */
  int ptr;          /* pointer in temporary list */
  type_t typ;       /* type of variables */
  int addr;         /* runtime address of variables */
  int length;       /* number of elements */
  int row_length;   /* number of elements per line */
  int i;            /* counter */

  aentry = listnext;    /* set aentry after last entry */
  ptr = firstconst - 4; /* start temporary table before constants */
  while (aentry > 0)
  {
    aentry = xdata[aentry - 1];           /* previous entry */
    if (ptr < listnext)
      serror (5);
    typ = xdata[aentry];

    if (!(typ & KCONST)) /* ignore procedures and functions */
      if ((typ & TYPEMASK) < LASTVARTYPE)
      {
        xdata[ptr]     = typ;
        xdata[ptr + 1] = xdata[aentry + 3]; /* runtime address */
        xdata[ptr + 2] = xdata[aentry + 4]; /* number of elements */
        xdata[ptr + 3] = xdata[aentry + 5]; /* number of elements per row */
        ptr -= 4;
      }
  }
  ptr += 4;                   /* back to first entry */

  while (ptr < firstconst)
  /* create variables = size + data */
  {
    /* read entry */
    typ = xdata[ptr]; /* type */
    addr = xdata[ptr + 1]; /* runtime addr of variable */
    length = xdata[ptr + 2]; /* number of elements (total) */
    row_length = xdata[ptr + 3]; /* number of elements (per line) */
    xdata[addr] = 0;
    if (typ & KREFERENCE)
      typ = DINTEGER; /* treat references like integers */
    if (row_length)
      xdata[addr - 2] = row_length; /* elements in array row */
    xdata[addr - 1] = length;     /* number of elements in array */

    for (i = 0; i < length; ++i)
      if ((typ & (TYPEMASK | KREFERENCE)) == DFLOAT)
        *(FLOATT FAR*) (&xdata[addr + FLOATSIZE * i]) = 0.0;
      else
        xdata[addr + i] = 0;
    ptr += 4;
  }
}

/*
   -------------  constants management ------------------
*/
void iconst (int i)
{
  --firstconst; /* reserve space */
  if ((firstconst < listnext) /* enough memory ? */
   || (firstconst < datanext))
    serror (33);
  xdata[firstconst] = i; /* store value */
}

void fconst (FLOATT f)
{
  firstconst -= FLOATSIZE; /* reserve necessary space */
  if ((firstconst < listnext) /* enough memory ?*/
   || (firstconst < datanext))
    serror (33);
  *((FLOATT FAR*) &xdata[firstconst]) = f; /* store value */
}
